/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 793 $ $LastChangedDate: 2002-08-31 21:59:39 +0200 (Sat, 31 Aug 2002) $
 *
 * Standard library for Gont, Hashtbl module implementation.
 *
 * Mutable hash table.
 */

open Hashtbl;

struct <'a>t {
	int size;
	int nelem;
	<<*(string, 'a)>list>Array::t tab;
}

<'a>t create(int sz)
{
	sz = prime_around(sz);

	let ini = fun (_) (Nil) in
	return {nelem = 0,
		size = sz,
		tab = Array::init(sz, ini)};
}

<'a>t copy(<'a>t src)
{
	return {nelem = src.nelem,
		size = src.size,
		tab = Array::copy(src.tab)};
}

int length(<'a>t h)
{
	return h.nelem;
}

void iter(<'a>t h, *(void (string, 'a)) f)
{
	void f2(*(string, 'a) p)
	{
		let (k, v) = p in
		f(k, v);
	}
	
	for (int i = 0; i < h.size; i++)
		List::iter(f2, Array::get(h.tab, i));
}

void check_rehash(<'a>t h)
{
	if (h.nelem < h.size)
		return;
		
	<'a>t n = create(prime_around(h.nelem * 4));

	let f = (fun void (string k, 'a v) (add(n, k, v))) in
	iter(h, f);

	h.tab = n.tab;
	h.size = n.size;
}

'a find_lst(<*(string, 'a)>list lst, string key)
{
	switch lst {
	case Nil:
		raise Not_found;
	case Cons((k, v), tl):
		if (k == key)
			return v;
		else
			return find_lst(tl, key);
	}
}

'a find(<'a>t h, string k)
{
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return find_lst(e, k);
}

<'a>list find_all_lst(<*(string, 'a)>list lst, string key)
{
	switch lst {
	case Nil:
		return Nil;
	case Cons((k, v), tl):
		if (k == key)
			return Cons(v, find_all_lst(tl, key));
		else
			return find_all_lst(tl, key);
	}
}

<'a>list find_all(<'a>t h, string k)
{
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return find_all_lst(e, k);
}

bool mem(<'a>t h, string k)
{
	bool xfind(<*(string, 'a)>list lst)
	{
		switch lst {
		case Nil:
			return false;
		case Cons((key, _), tl):
			if (k == key)
				return true;
			else
				return xfind(tl);
		}
	}
	
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return xfind(e);
}

void add(<'a>t h, string k, 'a v)
{
	h.nelem++;
	check_rehash(h);
	let pos = String::hash(k) % h.size in
	let e = Array::get(h.tab, pos) in
	Array::set(h.tab, pos, Cons((k, v), e));
}

void remove(<'a>t h, string k)
{
	<*(string, 'a)>list xfind(<*(string, 'a)>list lst)
	{
		switch lst {
		case Nil:
			raise Not_found;
		case Cons(hd, tl):
			let (key, _) = hd in
			if (k == key) {
				h.nelem--;
				return tl;
			} else {
				return Cons(hd, xfind(tl));
			}
		}
	}
	
	let idx = String::hash(k) % h.size in
	Array::set(h.tab, idx, xfind(Array::get(h.tab, idx)));
}

void remove_all(<'a>t h, string k)
{
	bool aint_our_key(*(string, 'a) p)
	{
		let (key, _) = p in 
		if (k != key) {
			h.nelem--;
			return true;
		} else {
			return false;
		}
	}
	
	let idx = String::hash(k) % h.size in
	let l = Array::get(h.tab, idx) in
	Array::set(h.tab, idx, List::filter(aint_our_key, l));
}

// TODO: can be faster
void replace(<'a>t h, string k, 'a v)
{
	remove_all(h, k);
	add(h, k, v);
}

def keys(h)
{
	_ ret = [];
	
	def add(k, _)
	{
		ret = Cons(k, ret);
	}

	iter(h, add);

	return ret;
}

def values(h)
{
	_ ret = [];
	
	def add(_, v)
	{
		ret = Cons(v, ret);
	}

	iter(h, add);

	return ret;
}
